<head>
	<title>Festival Class Hierachy</title>
</head>
<body background="http://www.cstr.ed.ac.uk/images/cstr_background.jpg" 
	bgcolor="#ffffff" link="#0000ff" 
	vlink="#dd0000" text="#000088" alink="9000ff">

<A HREF = "http://www.cstr.ed.ac.uk/">
   <IMG align=left BORDER=0 HSPACE=10 SRC="cstr.gif"></A> 
<A HREF="http://www.cstr.ed.ac.uk/projects/festival.html">
<IMG BORDER=0 ALIGN=right SRC="festival_tiny.jpg"></A>

<br>

<h1>
	Festival Documentation<br>
	<small><small><b>V1.2.5 December 1997</b></small></small>
</h1>
<br clear=left>
