/*************************************************************************/
/*                                                                       */
/*                Centre for Speech Technology Research                  */
/*                 (University of Edinburgh, UK) and                     */
/*                           Korin Richmond                              */
/*                         Copyright (c) 2002                            */
/*                         All Rights Reserved.                          */
/*                                                                       */
/*  Permission is hereby granted, free of charge, to use and distribute  */
/*  this software and its documentation without restriction, including   */
/*  without limitation the rights to use, copy, modify, merge, publish,  */
/*  distribute, sublicense, and/or sell copies of this work, and to      */
/*  permit persons to whom this work is furnished to do so, subject to   */
/*  the following conditions:                                            */
/*                                                                       */
/*   1. The code must retain the above copyright notice, this list of    */
/*      conditions and the following disclaimer.                         */
/*   2. Any modifications must be clearly marked as such.                */
/*   3. Original authors' names are not deleted.                         */
/*   4. The authors' names are not used to endorse or promote products   */
/*      derived from this software without specific prior written        */
/*      permission.                                                      */
/*                                                                       */
/*  THE UNIVERSITY OF EDINBURGH AND THE CONTRIBUTORS TO THIS WORK        */
/*  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING      */
/*  ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT   */
/*  SHALL THE UNIVERSITY OF EDINBURGH NOR THE CONTRIBUTORS BE LIABLE     */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES    */
/*  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN   */
/*  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,          */
/*  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF       */
/*  THIS SOFTWARE.                                                       */
/*                                                                       */
/*************************************************************************/
/*                                                                       */
/*                        Author: Korin Richmond                         */
/*                          Date: October 2002                           */
/* --------------------------------------------------------------------- */
/* name speaks for itself                                                */
/*                                                                       */
/*                                                                       */
/*                                                                       */
/*                                                                       */
/*                                                                       */
/*************************************************************************/

#include "ling_class/EST_Relation.h"
#include "ling_class/EST_Utterance.h"

#include "EST_JoinCost.h"
#include "EST_JoinCostCache.h"
#include "safety.h"
#include "ling_class/EST_Item.h"
#include "EST_FMatrix.h"

EST_JoinCost::~EST_JoinCost()
{
  int cclen = costCaches.length();
  for( int i=0; i<cclen; ++i )
    delete costCaches[i];
}

bool EST_JoinCost::computeAndCache( const EST_TList<EST_Item*> &list, bool verbose )
{
  // for later speed use the costCache's index as it's id
  unsigned int jccid = costCaches.length();
  costCaches.resize( jccid+1 ); //EST_TVector doesn't have push_back yet
  
  EST_JoinCostCache *jcc = new EST_JoinCostCache( jccid, list.length() );
  CHECK_PTR(jcc);

  costCaches[jccid] = jcc;  

  // this function adds the relevant indicies
  return jcc->computeAndCache( list, *this, verbose );
}




